# -*- coding: utf-8 -*-

# Copyright 2019 Michael V. Edmunds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

VERSION = (0, 3)
__version__ = ".".join(str(v) for v in VERSION)

NAME = "aws-cfn-ses-domain"
DESCRIPTION = "AWS CloudFormation resources for Amazon SES domain and email identities"
HOMEPAGE = "https://github.com/medmunds/aws-cfn-ses-domain"
AUTHOR = "Mike Edmunds"
AUTHOR_EMAIL = "medmunds@gmail.com"
LICENSE = "Apache License 2.0"

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Topic :: System :: Installation/Setup',
    'Topic :: System :: Systems Administration',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
]

KEYWORDS = [
    "AWS-CloudFormation",
    "CloudFormation",
    "CloudFormation-custom-resource",
    "CustomResource",
    "Amazon-SES",
    "SES",
    "Route53",
    "AWS-Lambda-Function",
]



def format_arn(partition=None, service=None, region=None, account=None,
               resource=None, resource_type=None, resource_name=None,
               defaults_from=None):
    """Return an ARN composed from the specified components.

    Provide either resource or both resource_type and resource_name.

    defaults_from can be an existing ARN, which will be used to fill in any
    missing components.

    See https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html
    for the format.
    """
    if resource is None and resource_type is not None:
        resource = f"{resource_type}/{resource_name}"
    if defaults_from is not None:
        try:
            _arn, _partition, _service, _region, _account, _resource = defaults_from.split(":")
        except (TypeError, ValueError):
            raise ValueError(f"Invalid ARN in defaults_from={defaults_from!r}")
        partition = partition if partition is not None else _partition
        service = service if service is not None else _service
        region = region if region is not None else _region
        account = account if account is not None else _account
        resource = resource if resource is not None else _resource

    return f"arn:{partition}:{service}:{region}:{account}:{resource}"


def to_bool(val):
    """Convert val to True or False.

    Converts 'true' (case-insensitive) and 1, '1', or True to True.
    Converts 'false', 'null' or 'none' (case-insensitive), the empty string '',
    and 0, '0', or False to False.
    Raises a ValueError for any other input.

    >>> to_bool('true')
    True
    >>> to_bool('False')
    False
    >>> to_bool(0)
    False
    >>> to_bool('0')
    False
    >>> to_bool(None)
    False
    >>> to_bool('yes')
    ValueError("Invalid boolean value 'yes'")
    """
    # (Loosely adapted from distutils.util.strtobool)
    strval = str(val).lower()
    if strval in ('true', '1'):
        return True
    elif strval in ('false', '0', 'null', 'none', ''):
        return False
    else:
        raise ValueError(f"Invalid boolean value {val!r}")



# cfnresponse (simplified for Python 3)
# adapted from https://github.com/jorgebastida/cfn-response
import json
import logging
from urllib.error import HTTPError
from urllib.request import Request, urlopen

logger = logging.getLogger()


SUCCESS = "SUCCESS"
FAILED = "FAILED"


def send(event, context, response_status, reason=None, response_data=None, physical_resource_id=None):
    response_body = json.dumps({
        "Status": response_status,
        "Reason": reason or "See the details in CloudWatch Log Stream: {}".format(context.log_stream_name),
        "PhysicalResourceId": physical_resource_id or context.log_stream_name,
        "StackId": event["StackId"],
        "RequestId": event["RequestId"],
        "LogicalResourceId": event["LogicalResourceId"],
        "Data": response_data or {}
    })
    logger.info("Sending response %r", response_body)

    request = Request(event["ResponseURL"], method="PUT",
                      data=response_body.encode("utf-8"),
                      headers={"Content-Type": ""})  # "application/json" will cause 403
    try:
        response = urlopen(request)
    except HTTPError as exc:
        logger.exception("Error sending response: code=%s", exc.code)
        return False
    else:
        logger.info("Successfully sent response: status=%s reason=%s", response.status, response.reason)
        return True



from setuptools import setup, find_packages
from os import path


here = path.abspath(path.dirname(__file__))
packages = find_packages(exclude=['build', 'contrib', 'docs', 'tests'])

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get package metadata from __about__.py (without importing the package)
about = {}
for package in packages:
    try:
        with open(path.join(here, package, '__about__.py')) as f:
            exec(f.read(), about)
    except FileNotFoundError:
        pass

setup(
    name=about["NAME"],
    version=about["__version__"],
    description=about["DESCRIPTION"],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=about["HOMEPAGE"],
    license=about["LICENSE"],
    author=about["AUTHOR"],
    author_email=about["AUTHOR_EMAIL"],
    classifiers=about["CLASSIFIERS"],
    keywords=" ".join(about["KEYWORDS"]),

    packages=packages,
    python_requires='>=3.6.0',
    install_requires=[
        # Requirements listed here will be bundled into the Lambda Function zip file.
        # Note that the AWS Lambda execution environment has boto3 pre-installed.
        # But if you need a newer version, use something like this to bundle boto3 too:
        # 'boto3>=1.12',
    ],
    extras_require={
        # use `pipenv install --dev` for dev requirements
        'test': ['boto3'],
    },
)



# Amazon SES email identity provisioning

import logging
import os

import boto3
from botocore.exceptions import BotoCoreError, ClientError

from .cfnresponse import FAILED, SUCCESS, send
from .utils import format_arn


logger = logging.getLogger()
logger.setLevel(os.getenv("LOG_LEVEL", "WARNING"))


DEFAULT_PROPERTIES = {
    "EmailAddress": "",
    "Region": os.getenv("AWS_REGION"),
}


def handle_email_identity_request(event, context):
    logger.info("Received event %r", event)

    properties = DEFAULT_PROPERTIES.copy()
    properties.update(event["ResourceProperties"])
    logger.info("Expanded properties to %r", properties)

    # Clean and validate inputs
    try:
        properties["EmailAddress"] = properties["EmailAddress"].strip()
    except (AttributeError, TypeError):
        pass
    email_address = properties["EmailAddress"]

    if not email_address:
        return send(event, context, FAILED,
                    reason="The 'EmailAddress' property is required.",
                    physical_resource_id="MISSING")

    ses = boto3.client("ses", region_name=properties["Region"])

    # Use an SES Identity ARN as the PhysicalResourceId - see:
    # https://docs.aws.amazon.com/IAM/latest/UserGuide/list_amazonses.html#amazonses-resources-for-iam-policies
    email_arn = format_arn(
        service="ses", region=properties["Region"],
        resource_type="identity", resource_name=email_address,
        defaults_from=event["StackId"])  # current stack's ARN has account and partition

    try:
        if event["RequestType"] == "Delete":
            response = ses.delete_identity(Identity=email_address)
            logger.info("SES:DeleteIdentity(Identity=%r) => %r", email_address, response)
        else:
            # Both Create and Update validate the new EmailAddress.
            # (For Update, the change in physical_resource_id will cause CloudFormation
            # to issue a Delete on the old EmailAddress after this request succeeds.)
            response = ses.verify_email_identity(EmailAddress=email_address)
            logger.info("SES:VerifyEmailIdentity(EmailAddress=%r) => %r", email_address, response)
    except (BotoCoreError, ClientError) as error:
        # for ClientError, might be helpful to look at error.response, too
        logger.exception("Error updating SES: %s", error)
        return send(event, context, FAILED,
                    reason=str(error), physical_resource_id=email_arn)

    outputs = {
        "Arn": email_arn,
        "EmailAddress": email_address,
        "Region": properties["Region"],
    }
    return send(event, context, SUCCESS,
                response_data=outputs, physical_resource_id=email_arn)



# AWS Lambda handler implementing Amazon SES domain identity provisioning
# for use as a CloudFormation CustomResource

import logging
import os

import boto3
from botocore.exceptions import BotoCoreError, ClientError

from .cfnresponse import FAILED, SUCCESS, send
from .utils import format_arn, to_bool

logger = logging.getLogger()
logger.setLevel(os.getenv("LOG_LEVEL", "WARNING"))


DEFAULT_PROPERTIES = {
    "Domain": "",
    "EnableSend": True,
    "EnableReceive": False,
    "MailFromSubdomain": "mail",
    "CustomDMARC": '"v=DMARC1; p=none; pct=100; sp=none; aspf=r;"',
    "TTL": "1800",
    "Region": os.getenv("AWS_REGION"),  # where the stack (lambda fn) is running
}
BOOLEAN_PROPERTIES = ("EnableSend", "EnableReceive")


def handle_domain_identity_request(event, context):
    logger.info("Received event %r", event)

    properties = DEFAULT_PROPERTIES.copy()
    properties.update(event["ResourceProperties"])
    logger.info("Expanded properties to %r", properties)

    # Clean and validate inputs
    try:
        properties["Domain"] = properties["Domain"].strip().rstrip(".")
    except (AttributeError, TypeError):
        pass
    domain = properties["Domain"]

    if not domain:
        return send(event, context, FAILED,
                    reason="The 'Domain' property is required.",
                    physical_resource_id="MISSING")

    # Use an SES Identity ARN as the PhysicalResourceId - see:
    # https://docs.aws.amazon.com/IAM/latest/UserGuide/list_amazonses.html#amazonses-resources-for-iam-policies
    domain_arn = format_arn(
        service="ses", region=properties["Region"],
        resource_type="identity", resource_name=domain,
        defaults_from=event["StackId"])  # current stack's ARN has account and partition

    for prop in BOOLEAN_PROPERTIES:
        # CloudFormation may convert YAML/JSON bools to strings, so reverse that
        # https://github.com/medmunds/aws-cfn-ses-domain/issues/10
        try:
            properties[prop] = to_bool(properties[prop])
        except ValueError:
            return send(event, context, FAILED,
                        reason=f"The '{prop}' property must be 'true' or 'false',"
                               f" not '{properties[prop]}'.",
                        physical_resource_id=domain_arn)

    if event["RequestType"] == "Delete" and event["PhysicalResourceId"] == domain:
        # v0.3 backwards compatibility:
        # Earlier versions used just the domain as the PhysicalResourceId.
        # When a CF update results in a new v0.3 id (ARN, rather than domain), CF will
        # automatically issue a Delete on the old id. We need to ignore that
        # request (or we'd incorrectly delete the domain we meant to provision).
        return send(event, context, SUCCESS,
                    response_data={"Domain": domain},
                    physical_resource_id=domain)

    if event["RequestType"] == "Delete":
        # Treat Delete as a request to disable both directions
        properties["EnableSend"] = False
        properties["EnableReceive"] = False

    # Update SES
    try:
        outputs = update_ses_domain_identity(domain, properties)
    except (BotoCoreError, ClientError) as error:
        # for ClientError, might be helpful to look at error.response, too
        logger.exception("Error updating SES: %s", error)
        return send(event, context, FAILED,
                    reason=str(error), physical_resource_id=domain_arn)

    # Determine required DNS
    properties.update(outputs)
    route53_records = generate_route53_records(properties)
    outputs.update({
        "Arn": domain_arn,
        "Domain": domain,
        "Region": properties["Region"],
        "Route53RecordSets": route53_records,
        "ZoneFileEntries": route53_to_zone_file(route53_records),
    })

    return send(event, context, SUCCESS,
                response_data=outputs, physical_resource_id=domain_arn)


def update_ses_domain_identity(domain, properties):
    """Handle SES (de-)provisioning for domain and returns dict of output info"""
    ses = boto3.client('ses', region_name=properties['Region'])

    outputs = {}
    enable_send = properties["EnableSend"]
    enable_receive = properties["EnableReceive"]

    if enable_send or enable_receive:
        response = ses.verify_domain_identity(Domain=domain)
        logger.info("SES:VerifyDomainIdentity(Domain=%r) => %r", domain, response)
        outputs["VerificationToken"] = response["VerificationToken"]
    else:
        # Neither send nor receive, so de-provision
        response = ses.delete_identity(Identity=domain)
        logger.info("SES:DeleteIdentity(Identity=%r) => %r", domain, response)

    if enable_send:
        response = ses.verify_domain_dkim(Domain=domain)
        logger.info("SES:VerifyDomainDKIM(Domain=%r) => %r", domain, response)
        # ??? ses.set_identity_dkim_enabled(Identity=domain, DkimEnabled=True)
        outputs["DkimTokens"] = response["DkimTokens"]

    if enable_send and properties["MailFromSubdomain"]:
        mail_from_domain = "{MailFromSubdomain}.{Domain}".format(**properties)
        outputs.update({
            "MailFromDomain": mail_from_domain,
            "MailFromMX": "feedback-smtp.{Region}.amazonses.com".format(**properties),
            "MailFromSPF": '"v=spf1 include:amazonses.com -all"',
        })
    else:
        # Disable custom Mail FROM domain.
        # (Could check first using ses.get_identity_mail_from_domain,
        # but clearing it doesn't cause an error even if not set/applicable.)
        mail_from_domain = ""
    response = ses.set_identity_mail_from_domain(Identity=domain, MailFromDomain=mail_from_domain)
    logger.info("SES:SetIdentityMailFromDomain(Domain=%r, MailFromDomain=%r) => %r",
                domain, mail_from_domain, response)

    if enable_send and properties["CustomDMARC"]:
        outputs["DMARC"] = properties["CustomDMARC"]

    if enable_receive:
        outputs.update({
            "ReceiveMX": "inbound-smtp.{Region}.amazonaws.com".format(**properties),
        })

    return outputs


def generate_route53_records(properties):
    """Return list of AWS::Route53::RecordSet resources required"""
    records = []

    if properties.get("VerificationToken"):
        records.append({
            "Name": "_amazonses.{Domain}.".format(**properties),
            "Type": "TXT",
            "ResourceRecords": ['"{VerificationToken}"'.format(**properties)]})

    if properties.get("DkimTokens"):
        records.extend([{
            "Name": "{token}._domainkey.{Domain}.".format(token=token, **properties),
            "Type": "CNAME",
            "ResourceRecords": ["{token}.dkim.amazonses.com.".format(token=token)],
        } for token in properties["DkimTokens"]])

    if properties.get("MailFromDomain"):
        if properties.get("MailFromMX"):
            records.append({
                "Name": "{MailFromDomain}.".format(**properties),
                "Type": "MX",
                "ResourceRecords": ["10 {MailFromMX}.".format(**properties)]})
        if properties.get("MailFromSPF"):
            records.append({
                "Name": "{MailFromDomain}.".format(**properties),
                "Type": "TXT",
                "ResourceRecords": [properties["MailFromSPF"]]})

    if properties.get("DMARC"):
        records.append({
            "Name": "_dmarc.{Domain}.".format(**properties),
            "Type": "TXT",
            "ResourceRecords": [properties["DMARC"]]})

    if properties.get("ReceiveMX"):
        records.append({
            "Name": "{Domain}.".format(**properties),
            "Type": "MX",
            "ResourceRecords": ["10 {ReceiveMX}.".format(**properties)]})

    for record in records:
        record["TTL"] = properties["TTL"]
    return records


def route53_to_zone_file(records):
    """Return a list of Zone File lines from a list of AWS::Route53::RecordSet"""
    max_name_len = max([len(record["Name"]) for record in records], default=1)
    return [
        "{name:{max_name_len}}\t{ttl}\tIN\t{type:5}\t{data}".format(
            name=record["Name"], max_name_len=max_name_len,
            ttl=record["TTL"], type=record["Type"],
            data=" ".join(record["ResourceRecords"]))
        for record in records]
